"""
A module to visualize matches.
"""
import sys, math

FONT_SIZE = 8
MAX_COLOR_VAL = 3
DARK = (20, 20, 20)

NOCOVER_SYMB = "*"


try:
    from fpdf import FPDF
except ImportError as exc:
    FPDF = None
    print(f"PDF generation disabled: {exc}", file=sys.stderr)


def floatRgb(mag, cmin, cmax):
    """
    Return a tuple of floats between 0 and 1 for the red, green and blue amplitudes.
    """

    try:
        # normalize to [0,1]
        x = float(mag - cmin) / float(cmax - cmin)
    except:
        # cmax = cmin
        x = 0.5
    blue = min((max((4 * (0.75 - x), 0.)), 1.))
    red = min((max((4 * (x - 0.25), 0.)), 1.))
    green = min((max((4 * math.fabs(x - 0.5) - 1., 0.)), 1.))
    return (red, green, blue)



def rgb(mag, cmin, cmax):
    """
    Return a tuple of integers to be used in AWT/Java plots.
    """
    mag = math.log(mag + 3)
    red, green, blue = floatRgb(mag, cmin, cmax)
    return (int(red * 255), int(green * 255), int(blue * 255))


def text_layout(full, trees, name, sequence, label, stream=sys.stdout):

    print(f">{name} (Mode={label})", file=stream)

    print(sequence, file=stream)

    line = [' '] * len(sequence)
    for index, tmp in enumerate(sequence):
        if not full.overlaps(index):
            line[index] = NOCOVER_SYMB

    print("".join(line), file=stream)

    for tree in trees:
        line = [' '] * len(sequence)
        for ival in tree:
            start, end, motif = ival.begin, ival.end, ival.data.pep.motif
            line[start:start + len(motif)] = motif

        print("".join(line), file=stream)

def save_data(matches, stream=sys.stdout):
    for m in matches:
        line = [ m.name, m.pep.motif, ".", m.start + 1, m.end, ".", m.pep.value, ".", f"Mode={m.pep.label}"]
        line = map(str, line)
        line = "\t".join(line)
        print(line, file=stream)


def get_canvas():
    """
    Returns the PDF canvas.
    """

    if not FPDF:
        return None

    pdf = FPDF(format="A4")
    pdf.add_page(orientation="L")
    pdf.set_font("Courier", size=FONT_SIZE)
    return pdf

def save_canvas(pdf, fname="output.pdf"):

    if pdf is None:
        return

    pdf.write(FONT_SIZE / 2, f"COLOR LEGEND:")
    pdf.ln()

    for value in range(1, 10):
        red, green, blue = rgb(value, 1, MAX_COLOR_VAL)
        pdf.set_text_color(red, green, blue)
        pdf.write(FONT_SIZE / 2, f"Value = {value}")
        pdf.ln()

    pdf.output(fname)

def pdf_layout(pdf, full, trees, name, sequence, label, file=sys.stdout):
    if pdf is None:
        return

    # Set dark color.
    pdf.set_text_color(*DARK)

    # Add the name of the sequence
    pdf.write(FONT_SIZE / 2, f">{name} (Mode={label})")
    pdf.ln()

    # Add the sequence
    pdf.write(FONT_SIZE / 2, sequence)
    pdf.ln()

    # Add the no coverage line
    line = [' '] * len(sequence)
    for index, tmp in enumerate(sequence):
        if not full.overlaps(index):
            line[index] = NOCOVER_SYMB
    line = "".join(line)

    pdf.write(FONT_SIZE / 2, line)
    pdf.ln()

    for tree in trees:
        last = 0
        for ival in sorted(tree):
            start, end, match = ival.begin, ival.end, ival.data
            skip = start - last
            if skip:
                empty = " " * skip
                pdf.write(FONT_SIZE / 2, empty)

            value = float(match.pep.value)
            red, green, blue = rgb(value, 1, MAX_COLOR_VAL)
            pdf.set_text_color(red, green, blue)
            pdf.write(FONT_SIZE / 2, match.pep.motif)
            pdf.set_text_color(*DARK)
            last = end

        pdf.ln()

    pdf.ln()


import plac, sys
from pathlib import Path
from . import parse, align, render
from pprint import pprint

def error(msg, stop=True):
    """
    Print error message.
    """
    print(msg, file=sys.stderr)
    if stop:
        sys.exit(1)

def warn(msg):
    """
    Print a warning message.
    """
    error(msg, stop=False)

@plac.opt('mass', "Mass-spec result file containing peptide sequences.")
@plac.opt('ref', "Reference file to match the peptides against.")
@plac.opt('pdf', "Output file for pdf file")
@plac.opt('txt', "Output file for text alignments")
@plac.opt('gff', "Output file as GFF data")
def run(mass, ref, pdf="output.pdf", txt="output.txt", gff="output.gff"):

    if not (mass or ref):
        parser = plac.parser_from(run)
        parser.print_help()
        sys.exit(1)

    # Will reuse this variable name but it makes for better parameter.
    pname = pdf

    # The labels
    labels = [ "POS", "NEG" ]

    fasta = parse.read_fasta(ref)

    pdf = render.get_canvas()

    dstream = open(gff, 'wt')
    tstream = open(txt, 'wt')

    for label in labels:

        peps = parse.read_peptides(fname=mass, label=label)

        for name, sequence in fasta.items():

            matches = align.find_matches(peps=peps, name=name, sequence=sequence)

            render.save_data(matches, stream=dstream)

            full, trees = align.find_layout(matches=matches)

            render.text_layout(full=full, trees=trees, name=name, sequence=sequence, label=label, stream=tstream)

            render.pdf_layout(pdf=pdf, full=full, trees=trees, name=name, sequence=sequence, label=label)


    if pname:
        render.save_canvas(pdf=pdf, fname=pname)


if __name__ == '__main__':
    plac.call(run)



import csv, sys


class Peptide(object):

    def __init__(self, motif, value, label):
        self.motif, self.value, self.label = motif, safe_float(value),  label

    def __repr__(self):
        return str(f"P:{self.__dict__}")

def safe_float(text):
    """
    Converts text to float. Returns 0 on invalid float.
    """
    try:
        return float(text)
    except ValueError as exc:
        return 0


def read_fasta(fname):
    """
    Parses a multi record fasta file.

    Returns a dictionary with sequence name as the key and sequence as value.
    """
    store = dict()

    header = None

    for line in open(fname):
        line = line.strip()

        if line.startswith(">"):
            # New fasta record
            header = line[1:].split()[0]
            store[header] = ""
        else:
            # Collect the fasta sequence
            store[header] += line

    return store

def read_peptides(fname, label=None):
    """
    Reads and parses an input file with measurements for each peptide.
    Returns a dictionary keyed by each individual pattern with the rest of the row as values for the key.
    """
    store = []

    stream = csv.reader(open(fname), delimiter="\t")

    # Skip header
    next(stream)

    # Apply label if needed
    if label:
        stream = filter(lambda x: x[2] == label, stream)

    # Iterate over the stream
    for row in stream:

        # Read the first three columns
        first, value, label = row[0:3]

        value = safe_float(value)

        peptides = first.split(";")

        for motif in peptides:
            motif = motif.strip()
            pep = Peptide(motif=motif, value=value, label=label)
            store.append(pep)

    return store

def print_locs(locs, stream=sys.stdout, sep=","):
    """
    Prints the locations as tabular file.
    """
    header = ["target", "peptide", "start", "end", "value", "label"]

    line = sep.join(header)

    print(line, file=stream)

    for row in locs:

        row = map(str, row)

        line = sep.join(row)

        print(line, file=stream)



from intervaltree import Interval, IntervalTree
from pprint import pprint
import sys

class Match(object):
    """
    Represents a match
    """
    def __init__(self, name, start, end, pep):
        self.name, self.start, self.end, self.pep = name, start, end, pep

    def __repr__(self):
        return str(f"M:{self.__dict__}")

def find_matches(peps, name, sequence):
    """
    Takes a list of peptides and matches them against each
    """
    matches = list()

    for step, tmp in enumerate(sequence):
        target = sequence[step:]
        for pep in peps:
            if target.startswith(pep.motif):
                start, end = step, step + len(pep.motif)
                match = Match(name=name, start=start, end=end, pep=pep)
                matches.append(match)

    return matches


def find_layout(matches, pad=1):
    """
    Finds the layout that can packs alignments into the fewest number of lines.
    """

    # The packing style depends on which intervals are added first.
    matches = sorted(matches, key=lambda m: (m.start, m.end))

    def find_tree(ival, trees, pad=pad):
        """
        Finds (or creates) an interval tree where the interval fits.
        """
        for tree in trees:
            if not tree.overlaps(ival.begin - pad, ival.end + pad):
                return tree

        tree = IntervalTree()
        trees.append(tree)
        return tree

    trees = []
    full = IntervalTree()
    for match in matches:
        ival = Interval(match.start, match.end, data=match)
        tree = find_tree(ival, trees=trees)
        tree.add(ival)
        full.add(ival)

    return full, trees




import setuptools
import pepaln

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="pepaln",
    version=pepaln.VERSION,
    author="Istvan Albert",
    author_email="istvan.albert@gmail.com",
    description="Peptide Matcher ",
    long_description=long_description,
    long_description_content_type="text/markdown",
    #url="https://github.com/biostars/biostar-handbook-code",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        'plac>=1.3',
        'intervaltree',
        'FPDF',
    ],

    python_requires='>=3.6',

)
